/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.util.Arrays;
import java.util.ConcurrentModificationException;
import java.util.List;
import java.util.logging.Logger;

import org.eiichiro.acidhouse.Aggregation;
import org.eiichiro.acidhouse.Coordinator;
import org.eiichiro.acidhouse.DatastoreSession;
import org.eiichiro.acidhouse.Delete;
import org.eiichiro.acidhouse.EntityExistsException;
import org.eiichiro.acidhouse.GetList;
import org.eiichiro.acidhouse.GetScalar;
import org.eiichiro.acidhouse.IndoubtException;
import org.eiichiro.acidhouse.ResourceManager;
import org.eiichiro.acidhouse.Update;

import com.google.appengine.api.datastore.DatastoreService;
import com.google.appengine.api.datastore.DatastoreServiceFactory;
import com.google.appengine.api.datastore.Transaction;

/**
 * {@code AppEngineDatastoreSession} is a App Engine Low-level Datastore API 
 * based {@code DatastoreSession} implementation.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class AppEngineDatastoreSession implements DatastoreSession {

	private final Logger logger = Logger.getLogger(getClass().getName());
	
	private final AppEngineDatastoreService datastore;
	
	public AppEngineDatastoreService getAppEngineDatastoreService() {
		return datastore;
	}

	private ThreadLocal<Coordinator> coordinator = new ThreadLocal<Coordinator>();
	
	/**
	 * Constructs a new {@code AppEngineDatastoreSession}.
	 * This App Engine Low-level Datastore API based {@code DatastoreSession} 
	 * implementation is called "Strong session". This implementation supports 
	 * the strong consistency with atomic commitment in multiple entity groups 
	 * and consistent read.
	 * And {@code AppEngineLooseDatastoreSession} supports "Loose session". See 
	 * {@link AppEngineLooseDatastoreSession}.
	 * 
	 * @see AppEngineLooseDatastoreSession
	 */
	public AppEngineDatastoreSession() {
		DatastoreService datastore = DatastoreServiceFactory.getDatastoreService();
		this.datastore = new AppEngineDatastoreService(datastore);
		logger.info("Datastore \"Strong\" Session started");
	}
	
	/**
	 * Begins transaction with {@code AppEngineCoordinator}.
	 * You have to do transactional operation after calling this method.
	 */
	@Override
	public void begin() {
		if (isTransactional()) {
			throw new IllegalStateException("Transaction has already started ["
					+ coordinator.get().getCurrentTransaction().getId() + "]");
		}
		
		coordinator.set(new AppEngineCoordinator(datastore));
		logger.info("Transaction ["
				+ coordinator.get().getCurrentTransaction().getId()
				+ "] started");
	}

	/**
	 * Commits current transaction with {@code AppEngineCoordinator}.
	 * You have to invoke this method after calling {@code #begin()} method.
	 */
	@Override
	public void commit() throws IndoubtException {
		if (!isTransactional()) {
			throw new IllegalStateException("Transaction has not started yet");
		}
		
		coordinator.get().commit();
		logger.info("Transaction ["
				+ coordinator.get().getCurrentTransaction().getId()
				+ "] committed");
		coordinator.remove();
	}

	/**
	 * Rolls back current transaction with {@code AppEngineCoordinator}.
	 * You have to invoke this method after calling {@code #begin()} method.
	 */
	@Override
	public void rollback() {
		if (!isTransactional()) {
			throw new IllegalStateException("Transaction has not started yet");
		}
		
		coordinator.get().rollback();
		coordinator.remove();
	}

	/**
	 * Returns entity instance corresponding to the specified key with 
	 * {@code AppEngineResourceManager}. This method returns {@code null} if the 
	 * entity is not found.
	 * 
	 * @param <E> The entity type.
	 * @param clazz The entity type that you want to get.
	 * @param id The id corresponding to the entity that you want to get.
	 * @return The entity instance corresponding to the specified key.
	 * @throws ConcurrentModificationException If the entity corresponding to 
	 * the specified key is being modified by the other transaction.
	 * @see AppEngineResourceManager#get(Class, Object)
	 */
	@Override
	public <E> E get(Class<E> clazz, Object id) throws ConcurrentModificationException {
		if (clazz == null) {
			throw new IllegalArgumentException("'clazz' must not be [" + clazz + "]");
		}
		
		if (id == null) {
			throw new IllegalArgumentException("'id' must not be [" + id + "]");
		}
		
		if (!isTransactional()) {
			ResourceManager<Transaction> manager = new AppEngineResourceManager(datastore);
			return manager.get(clazz, id);
		} else {
			return coordinator.get().get(clazz, id);
		}
	}
	
	/**
	 * Puts the specified entity instance into App Engine Datastore newly.
	 * This method must be invoked under a transaction. This method throws 
	 * {@code EntityExistsException} if the entity that has the same key as 
	 * the specified entity has been already stored.
	 * 
	 * @param entity The entity instance to be put into App Engine Datastore.
	 * @throws EntityExistsException If the entity that has the same key as the 
	 * specified entity has already existed.
	 */
	@Override
	public void put(Object entity) throws EntityExistsException {
		if (entity == null) {
			throw new IllegalArgumentException("'entity' must not be [" + entity + "]");
		}
		
		assertTransactional();
		coordinator.get().put(entity);
	}
	
	/**
	 * Updates entity with the specified entity instance with 
	 * {@code AppEngineCoordinator}.
	 * This method must be invoked under a transaction. 
	 * 
	 * @param entity The entity instance to be updated.
	 */
	@Override
	public void update(Object entity) {
		if (entity == null) {
			throw new IllegalArgumentException("'entity' must not be [" + entity + "]");
		}
		
		assertTransactional();
		coordinator.get().update(entity);
	}
	
	private void assertTransactional() {
		if (!isTransactional()) {
			throw new IllegalStateException("Transaction has not started: "
					+ "Must call DatastoreSession#beginTransaction() "
					+ "before doing this");
		}
	}
	
	private boolean isTransactional() {
		return (coordinator.get() != null);
	}

	/**
	 * Deletes the specified entity from App Engine Datastore with 
	 * {@code AppEngineCoordinator}.
	 * This method must be invoked under a transaction.
	 * 
	 * @param entity The entity to be deleted.
	 */
	@Override
	public void delete(Object entity) {
		if (entity == null) {
			throw new IllegalArgumentException("'entity' must not be [" + entity + "]");
		}
		
		assertTransactional();
		coordinator.get().delete(entity);
	}

	/** Closes this session. */
	@Override
	public void close() {
		if (coordinator.get() != null) {
			coordinator.get().rollback();
		}
		
		coordinator.remove();
	}

	/**
	 * Returns {@code Delete} based on App Engine Low-level Datastore API for 
	 * the specified entity class.
	 * This method is the entry point for deleting Command Builder API.
	 * 
	 * @param <E> The type of entity deleted by this {@code Delete}.
	 * @param clazz The type of entity deleted by this {@code Delete}.
	 * @return The {@code Delete} for the specified entity class.
	 */
	@Override
	public <E> Delete<E> delete(Class<E> clazz) {
		return new AppEngineDelete<E>(clazz, this);
	}

	/**
	 * Returns {@code GetList} based on App Engine Low-level Datastore API for 
	 * the specified entity class.
	 * This method is the entry point for getting-list Command Builder API.
	 * 
	 * @param <E> The entity type to get with this {@code GetList}.
	 * @param clazz The entity type to get with this {@code GetList}. 
	 * @return {@code GetList} for the specified entity class.
	 * @see GetList
	 */
	@Override
	public <E> GetList<E> get(Class<E> clazz) {
		return new AppEngineGetList<E>(clazz, this);
	}

	/**
	 * Returns {@code GetScalar} based on App Engine Low-level Datastore API for 
	 * this specified {@code Aggregation}.
	 * This method is the entry point for aggregation Command Builder API.
	 * 
	 * @param <R> The property value type to aggregate with this 
	 * {@code GetScalar}.
	 * @param aggregation The {@code Aggregation} this method executes.
	 * @return {@code GetScalar} for the specified {@code Aggregation} 
	 * instance.
	 */
	@Override
	public <E, R> GetScalar<E, R> get(Aggregation<R> aggregation) {
		return new AppEngineGetScalar<E, R>(aggregation, this);
	}

	/**
	 * Returns {@code Update} based on App Engine Low-level Datastore API for 
	 * the specified entity class.
	 * This method is the entry point for updating Command Builder API.
	 * 
	 * @param <E> The type of entity updated by this {@code Update}.
	 * @param clazz The type of entity updated by this {@code Update}.
	 * @return The {@code Update} for the specified entity class.
	 */
	@Override
	public <E> Update<E> update(Class<E> clazz) {
		return new AppEngineUpdate<E>(clazz, this);
	}

	/**
	 * Deletes the specified entities from App Engine Datastore. 
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of deleted entity. 
	 * @param entities The entities to be deleted.
	 */
	@Override
	public <E> void delete(E... entities) {
		delete(Arrays.asList(entities));
	}

	/**
	 * Deletes the specified entity list from App Engine Datastore. 
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of deleted entity. 
	 * @param entities The entity list to be deleted.
	 */
	@Override
	public <E> void delete(List<E> entities) {
		if (entities == null) {
			throw new IllegalArgumentException("'entities' must not be ["
					+ entities + "]");
		}
		
		for (E entity : entities) {
			delete(entity);
		}
	}

	/**
	 * Puts the specified entity instances into App Engine Datastore newly.
	 * This method must be invoked under the transactional session. This method 
	 * throws {@code EntityExistsException} if the entity that has the same key 
	 * as the specified entity has been already stored.
	 * 
	 * @param <E> The type of entity instances to be put by this operation.
	 * @param entities The entity instances that you want to put into App Engine 
	 * Datastore.
	 * @throws EntityExistsException If the entity that has the same key as the 
	 * specified entity has been already existed.
	 */
	@Override
	public <E> void put(E... entities) throws EntityExistsException {
		put(Arrays.asList(entities));
	}

	/**
	 * Puts the specified entity list into App Engine Datastore newly.
	 * This method must be invoked under the transactional session. This method 
	 * throws {@code EntityExistsException} if the entity that has the same key 
	 * as the specified entity has been already stored.
	 * 
	 * @param <E> The type of entity instances to be put by this operation.
	 * @param entities The entity instances that you want to put into 
	 * App Engine Datastore as a {@code List} view.
	 * @throws EntityExistsException If the entity that has the same key as the 
	 * specified entity has been already existed.
	 */
	@Override
	public <E> void put(List<E> entities) throws EntityExistsException {
		if (entities == null) {
			throw new IllegalArgumentException("'entities' must not be ["
					+ entities + "]");
		}
		
		for (E entity : entities) {
			put(entity);
		}
	}

	/**
	 * Updates entities with the specified entity instances.
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of updated entity.
	 * @param entities The entities to be updated.
	 */
	@Override
	public <E> void update(E... entities) {
		update(Arrays.asList(entities));
	}

	/**
	 * Updates entities states with the specified entity instance list.
	 * This method must be invoked under the transaction.
	 * 
	 * @param <E> Type of updated entity.
	 * @param entities The entity list to be updated.
	 */
	@Override
	public <E> void update(List<E> entities) {
		if (entities == null) {
			throw new IllegalArgumentException("'entities' must not be ["
					+ entities + "]");
		}
		
		for (E entity : entities) {
			update(entity);
		}
	}
	
}
